"""
CBMPy: fluxmodules decomposiiton module
=====================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Arne C. Reimers
Contact email: arne.c.reimers@gmail.com
"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

import numpy

from .matroid import fromMatrix
from .random_color import generate_new_color


class Edge:
    """ Every edge connects two vertices """

    def __init__(self, a, b):
        """ initialize edge connecting vertices a and b """
        self.a = a
        self.b = b

    def getOther(self, vertex):
        """ give the other vertex then the specified one

        The specified vertex must be one of this edges vertices
        """
        if self.a == vertex:
            return self.b
        else:
            assert self.b == vertex, 'vertex must be a vertex of this edge'
            return self.a


class AbstractVertex:
    """ A vertex in a graph. It can be a leaf or an inner node """

    maxId = 0
    """ for giving unique ids for hashing """

    def __init__(self):
        self.__id = AbstractVertex.maxId
        AbstractVertex.maxId += 1
        pass

    def id(self):
        return self.__id

    def __hash__(self):
        return self.__id

    def __eq__(self, other):
        """ two abstract vertices are only the same if they are identical """
        return self is other

    def label(self):
        print("subclasses must override this")

    def replaceAdjacent(self, old, new):
        """ replace the adjacent node old by new """
        print('subclasses must override this')
        assert False

    def getAdjacent(self):
        """ list adjacent nodes """
        print('subclasses must override this')
        assert False

    def getLeaves(self, fromVertex):
        """ enumerates all the leaves of this vertex that are reachable without
        using fromVertex

        fromVertex must be a vertex adjacent to this vertex
        """
        print('subclasses must override this')
        assert False


class Vertex(AbstractVertex):
    """ represents an inner vertex (i.e., not a leaf)

    Every Vertex contains a map (edges) of adjacent edges resp. vertices.
    Each Edge has a list of elements associated to it (the interface).
    Each Vertex stores a matroid, whoose elements are contributed by the edges.

    The map edges has the following form:
    The keys are the incident vertices.
    The values are the list of elements contributed by the edge
    """

    def __init__(self):
        AbstractVertex.__init__(self)
        self.edges = {}

    def __repr__(self):
        return "N%s" % self.id()

    def __str__(self):
        return "N%s" % self.id()

    def label(self):
        return self.__str__()

    def setUpRoot(self, matroid):
        """ set ups this vertex as the root node for the given matroid.

        This means we add leaves for all the elements of the matroid to this
        vertex
        """
        for e in matroid.elems:
            l = Leaf(e, self)
            self.edges[l] = [e]
        self.matroid = matroid

    def splitOff(self, adjacent):
        """ splits of a list of adjacent vertices.

        adjacent must be a list of vertices that are adjacent to this node.
        The ist adjacent must contain more than one element
        """
        assert len(adjacent) >= 2, 'splitting off would have no effect'

        # the next assert may be useful, but it can make things also much
        # more complicated and it does not really hurt the branch decompositon
        #assert len(adjacent) <= len(self.edges) <= 2

        # collect elements to split off
        toSplit = set([])
        for a in adjacent:
            toSplit.update(self.edges[a])
        # find a basis of self.matroid such that the only basis elements of
        # toSplit are used by non-basis elements of toSplit

        # in the following loop, we modify the set of non-basis elements,
        # so first copy it to avoid concurrent modification errors
        for e in list(self.matroid.nbasis):
            if e not in toSplit:
                c = self.matroid.fundamentalCircuit(e)
                intersection = c.intersection(toSplit)
                if len(intersection) >= 1:
                    # put e into the basis
                    # and one element of the intersection into nbasis
                    # note that the new non-basis element is in toSplit, so we
                    # don't have to do anything with it
                    self.matroid.exchange(e, intersection.pop())

        # our matroid is now in the desired form:
        # no fundamental circuit of a nbasis element outside toSplit uses basis
        # elements of toSplit
        # and no cofundamental circuit  of a basis element in toSplit uses nbasis
        # elements outside of toSplit

        # create new vertex
        vsplit = Vertex()
        for a in adjacent:
            # move edge from self to vsplit
            vsplit.edges[a] = self.edges[a]
            # delete adjacency of a
            del self.edges[a]
            a.replaceAdjacent(self, vsplit)

        # the dual matroid is of the desired form for the complement of toSplit,
        # which we will use to simplify the interface of vsplit
        vsplit.matroid = self.matroid.dual()

        # reduce representation of toSplit for self
        interface = self.__simplifyInterface(toSplit)
        self.edges[vsplit] = interface  # add vsplit as neighbour

        if len(adjacent) >= 3 or True:
            # reduce representation of the complement of toSplit for vsplit
            co_toSplit = set(vsplit.matroid.elems) - toSplit
            cointerface = vsplit.__simplifyInterface(co_toSplit)
            vsplit.edges[self] = cointerface  # add self as neighbour to vsplit
            # in the end we actually want to have the primal matroid stored
            vsplit.matroid = vsplit.matroid.dual()
        else:
            # vsplit is a node with 3 vertices, we will probably not touch it
            # again, so be lazy and do not calculate the cointerface
            vsplit.edges[self] = len(interface) * [None]
            vsplit.matroid = None

        # return the new vertex, maybe we want to do something with it
        return vsplit

    def replaceAdjacent(self, old, new):
        assert old in self.edges, 'old must be an adjacent vertex'
        # copy interface
        self.edges[new] = self.edges[old]
        # delete old edge
        del self.edges[old]

    def __simplifyInterface(self, toSplit):
        """ reduces the interface of the elements to split off

        For this method we assume:
        No fundamental circuit of a non-basis element outside of toSplit must
        contain a basis element inside toSplit
        """
        assert self.__checkSplit(toSplit), 'basis has wrong form'

        self.matroid = self.matroid.contract(
            toSplit.intersection(self.matroid.basis)
        )
        # this gives us an interface of the split
        interface = []
        for i in self.matroid.nbasis:
            if i in toSplit:
                interface.append(i)
        # however, we want to simplify this interface
        # it should only have as many elements as the connectivity is

        # take the reduced representation matrix of the non-basis elements in
        # toSplit in self.matroid and interpret it as a direct representation
        # matrix and find a basis
        toDelete = [i for i in self.matroid.nbasis if i not in toSplit]
        # we only delete non-basis elements, so delete does not mess with basis
        submat = self.matroid.delete(toDelete)
        mred = fromMatrix(submat.rep, interface)
        # the basis elements of mred represent the interface
        # so we delete all the other elements
        toDelete = [i for i in interface if i in mred.nbasis]
        self.matroid = self.matroid.delete(toDelete)

        return mred.basis

    def __checkSplit(self, toSplit):
        for e in list(self.matroid.nbasis):
            if e not in toSplit:
                c = self.matroid.fundamentalCircuit(e)
                if not c.isdisjoint(toSplit):
                    return False
        return True

    def getAdjacent(self):
        return frozenset(self.edges.keys())

    def getLeaves(self, fromVertex):
        """ enumerates all the leaves of this vertex that are reachable without
        using fromVertex

        If fromVertex is not a vertex incident to this vertex, then all leaves
        are enumerated.
        """
        leaves = set()
        for v in self.edges.keys():
            if v != fromVertex:
                leaves.update(v.getLeaves(self))
        return leaves

    def isSimple(self):
        if len(self.matroid.elems) == 0:
            return True
        # check if its parallel elements
        parallel = self.matroid.parallel()
        if len(parallel) == 1:
            par = parallel.pop()
            if len(par) == len(self.matroid.elems):
                assert par == frozenset(self.matroid.elems)
                return True

        # check if its cpparallel elements
        coparallel = self.matroid.coparallel()
        if len(coparallel) == 1:
            copar = coparallel.pop()
            if len(copar) == len(self.matroid.elems):
                assert copar == frozenset(self.matroid.elems)
                return True

        return False

class Leaf(AbstractVertex):

    def __init__(self, element, adjacent):
        """ creates leaf representing an matroid element

        element is the matroid element to represent
        adjacent is the vertex to which this leaf is added
        """
        AbstractVertex.__init__(self)
        self.elem = element
        self.adj = adjacent

    def __repr__(self):
        return "Leaf%s <%s>" % (self.id(), self.elem)

    def __str__(self):
        return "Leaf%s <%s>" % (self.id(), self.elem)

    def label(self):
        return self.elem.__str__()

    def replaceAdjacent(self, old, new):
        assert self.adj == old, 'old must be single adjacent node of this leaf'
        self.adj = new

    def getAdjacent(self):
        return frozenset([self.adj])

    def getLeaves(self, fromVertex):
        assert self.adj == fromVertex, 'fromEdge must be from the leaf'
        return {self.elem}


class Decomposition:
    """ Hierarchical branch decomposition

    Actually can also represent a partial branch decomposition as intermediate
    results.

    It has the following fields:
    matroid - the matroid of this branch decomposition
    root - a vertex of the graph representing the branch decomposition

    TODO: more details
    """

    def __init__(self, matroid):
        """ initializes as star branch decomposition """
        self.matroid = matroid
        self.root = Vertex()
        self.root.setUpRoot(matroid)

    def listNonLeaves(self):
        """ list all non-leaf vertices """
        return self.listNonLeavesRec(self.root, None)

    def listNonLeavesRec(self, vertex, source):
        if isinstance(vertex, Leaf):
            return set()

        result = {vertex}
        for v in vertex.getAdjacent():
            if v != source:
                result.update(self.listNonLeavesRec(v, vertex))
        return result

    def getVertex(self, vertexid):
        """ returns the vertex with the given id

        If no vertex with the given id exists, None is returned.
        """
        for v in self.listNonLeaves():
            if v.id() == vertexid:
                return v
        return None

    def splitZeroModules(self, vertex):
        """ branch given vertex into 0 modules

        Requires that all adjacent vertices have at most 1-dimensional interface
        """
        lookUpAdjacent = {}
        for adjacent, elems in vertex.edges.items():
            assert len(elems) <= 1
            for e in elems:
                lookUpAdjacent[e] = adjacent

        modules = vertex.matroid.findModules()
        if len(modules) >= 2:  # otherwise, splitting helps not
            for m in modules:
                if len(m.elems) > 1:  # don't do anything for loops or coloops
                    toSplit = set([])
                    for e in m.elems:
                        toSplit.add(lookUpAdjacent[e])

                    vertex.splitOff(toSplit)
                else:
                    # for loops and coloops we have to fix the interface
                    e = m.elems[0]
                    v = lookUpAdjacent[e]
                    vertex.edges[v] = []
                    vertex.matroid = vertex.matroid.delete([e])

    def splitSimple(self):
        """ split all vertices into 0-modules and 1-modules """
        for v in self.listNonLeaves():
            self.splitZeroModules(v)

        for v in self.listNonLeaves():
            self.splitOneModules(v)

    def splitOneModules(self, vertex):
        """ branch given vertex into 1 modules

        TODO:we have to deal with grouped elements
        """
        # first try to group parallel and coparallel elements
        didSomething = True
        while didSomething:
            didSomething = False
            # create lookup table for adjacent vertices
            # the matroid changes, so we update it for each iteration
            lookUpAdjacent = {}
            for adjacent, elems in vertex.edges.items():
                for e in elems:
                    lookUpAdjacent[e] = adjacent

            # find parallel elements
            parallel = vertex.matroid.parallel()
            print("%s parallel: %s" % (vertex, parallel))
            for group in parallel:
                assert len(group) >= 2, 'groups must be non-trivial'
                # we split vertices not elements, so get the vertices
                toSplit = set()
                for e in group:
                    v = lookUpAdjacent[e]
                    if len(vertex.edges[v]) <= 1:
                        toSplit.add(v)
                if (len(toSplit) >= 2) and (len(toSplit) <= len(vertex.edges) - 2):
                    vertex.splitOff(toSplit)
                    didSomething = True

            # update lookup table if we changed something
            if didSomething:
                lookUpAdjacent = {}
                for adjacent, elems in vertex.edges.items():
                    for e in elems:
                        lookUpAdjacent[e] = adjacent

            # find coparallel elements
            coparallel = vertex.matroid.coparallel()
            print("%s coparallel: %s" % (vertex, coparallel))
            for group in coparallel:
                assert len(group) >= 2, 'groups must be non-trivial'
                # we split vertices not elements, so get the vertices
                toSplit = set()
                for e in group:
                    v = lookUpAdjacent[e]
                    if len(vertex.edges[v]) <= 1:
                        toSplit.add(v)
                if (len(toSplit) >= 2) and (len(toSplit) <= len(vertex.edges) - 2):
                    vertex.splitOff(toSplit)
                    didSomething = True

        print('TODO: more complicated 1-modules')

    def makeSubcubic(self):
        """ split all nodes that have degree >= 4 in an arbitrary manner."""
        for v in self.listNonLeaves():
            assert isinstance(v, Vertex)
            while len(v.getAdjacent()) >= 4:
                # split off two arbitrary nodes
                adj = v.getAdjacent()
                aiter = adj.__iter__()
                v.splitOff([aiter.next(), aiter.next()])

    def isParallelSeries(self):
        """ checks if all nodes have only parallel or only coparallel elems.

        This does not check if the matroid is parallel-series.
        """
        for v in self.listNonLeaves():
#             ok = False
#             # check if its parallel elements
#             parallel = v.matroid.parallel()
#             if len(parallel) == 1:
#                 par = parallel.pop()
#                 if len(par) == len(v.matroid.elems):
#                     assert par == frozenset(v.matroid.elems)
#                     ok = True
#
#             # check if its cpparallel elements
#             coparallel = v.matroid.coparallel()
#             if len(coparallel) == 1:
#                 copar = coparallel.pop()
#                 if len(copar) == len(v.matroid.elems):
#                     assert copar == frozenset(v.matroid.elems)
#                     ok = True
#
#             if not ok:
            if not v.isSimple():
                return False

        # all nodes are either parallel or coparallel
        return True

    def verifyFullyBranched(self):
        """ verifies that this branch decomposition is fully branched

        The branch decomposition is considered fully branched if all vertices
        satisfy:
        - is degree 3 or
        - all elements are parallel
        - all elements are coparallel

        If successful, True is returned.
        Otherwise, a NotFullyBranchedException is raised
        """
        for v in self.listNonLeaves():
            ok = False
            if len(v.edges) <= 3:
                ok = True

            if not ok:
                if v.isSimple():
                    ok = True
#                 # check if its parallel elements
#                 parallel = v.matroid.parallel()
#                 if len(parallel) == 1:
#                     par = parallel.pop()
#                     if len(par) == len(v.matroid.elems):
#                         assert par == frozenset(v.matroid.elems)
#                         ok = True
#
#             if not ok:
#                 # check if its cpparallel elements
#                 coparallel = v.matroid.coparallel()
#                 if len(coparallel) == 1:
#                     copar = coparallel.pop()
#                     if len(copar) == len(v.matroid.elems):
#                         assert copar == frozenset(v.matroid.elems)
#                         ok = True

            if not ok:
                raise NotFullyBranchedException(v)
                return False

        # all nodes are either parallel or coparallel
        return True

    def verifyEdgeWidth(self):
        """ verifies if the edge with really coincides with the interfaces.

        In case of success, true is returned.
        Otherwise an EdgeWidthException is raised.
        """
        for v in self.listNonLeaves():
            for (w, interface) in v.edges.items():
                sep = w.getLeaves(v)
                print(sep)
                conn = self.matroid.connectivity(sep)
                if len(interface) != conn:
                    raise EdgeWidthException(v, w, conn)
                    return False
        return True

    def getWidth(self):
        """ computes the width of this branch decomposition.
        """
        width = 0
        for v in self.listNonLeaves():
            for (w, interface) in v.edges.items():
                width = max(width, len(interface))
        return width

    def poolmanMethod(self, vertex, similarity=None):
        """ decomposes the matroid using the method suggested by Poolman et al.

        This method iteratively groups these elements together that are most
        similar. Similarity by sets of nodes is done via the mean.
        For more details see doi:10.1016/j.jtbi.2007.08.005

        This method takes a similarity matrix as input, which allows it to
        also apply the method to other similarity matrices then those suggested
        in the original work.
        If no similarity matrix is given, then a row similarity matrix is used,
        which corresponds to the matrix suggested by Poolman et al.
        """

        if similarity == None:
            similarity = self.matroid.similarityMatrix("rows")

        # make a copy so that we don't make a mess with the input
        sim = similarity.matrix.copy()

        if set(similarity.elems) == set(vertex.matroid.elems):
            # similarity matrix is directly for the vertex matroid

            # merge columns representing vertices
            # preallocate, so we don't get indexing problems
            vertexIndices = len(similarity.elems) * [None]
            for w, elems in vertex.edges.items():
                # typically len(elems) == 1
                sumrow = numpy.zeros(len(similarity.elems))
                i = -1
                for e in elems:
                    i = similarity.elems.index(e)
                    sumrow += sim[i, :]
#                    sumrow = numpy.maximum(sumrow, sim[i,:])
                    sim[i, :] = -1  # does not overwrite important entries
                    sim[:, i] = -1  # does not overwrite important entries
                    sumrow[i] = -1  # make sure the value is negative

                if len(elems) >= 1:
                    assert i >= 0
                    sumrow /= len(elems)
                    sim[i, :] = sumrow
                    sim[:, i] = sumrow
                    vertexIndices[i] = (w, len(elems))
        else:
            assert set(similarity.elems) == set(self.matroid.elems)
            # similarity is of the whole-network matroid

            # merge elements in adjacent nodes
            # preallocate, so we don't get indexing problems
            vertexIndices = len(similarity.elems) * [None]
            adj = vertex.getAdjacent()
            for a in adj:
                tomerge = a.getLeaves(vertex)
                sumrow = numpy.zeros(len(similarity.elems))
                i = -1
                for e in tomerge:
                    i = similarity.elems.index(e)
                    sumrow += sim[i, :]
#                    sumrow = numpy.maximum(sumrow, sim[i,:])
                    sim[i, :] = -1  # does not overwrite important entries
                    sim[:, i] = -1  # does not overwrite important entries
                    sumrow[i] = -1  # make sure the value is negative

                if len(tomerge) >= 1:
                    assert i >= 0
                    sumrow /= len(tomerge)
                    sim[i, :] = sumrow
                    sim[:, i] = sumrow
                    vertexIndices[i] = (a, len(tomerge))

        # now we can start with the actual algorithm
        while len(vertex.edges) >= 4:
            (i, j) = numpy.unravel_index(sim.argmax(), sim.shape)
            assert i != j
            assert sim[i, j] >= 0
            assert vertexIndices[i] != None
            assert vertexIndices[j] != None
            (vi, ni) = vertexIndices[i]
            (vj, nj) = vertexIndices[j]
            # merge i and j, because they are most similar
            vsplit = vertex.splitOff([vi, vj])
            # create entry for vertex in similarity matrix
            newrow = (ni * sim[i, :] + nj * sim[j, :]) / (ni + nj)
#            newrow = numpy.maximum(sim[j,:], sim[i,:])
            sim[i, :] = newrow
            sim[:, i] = newrow
            sim[j, :] = -1
            sim[:, j] = -1
            sim[i, i] = -1
            vertexIndices[i] = (vsplit, ni + nj)
            vertexIndices[j] = None

    def makeGraphViz(self, cmod=None):
        """ create graphviz string to draw this decomposition """
        out = "graph G {\n"
        found = set()
        for v in self.listNonLeaves():
            for (w, e) in v.edges.items():
                if w not in found:
                    out += v.label() + " -- " + w.label()
                    out += " [label = " + len(e).__str__() + "]\n"
            found.add(v)

        # color leaves according to subsystems
        if cmod != None:
            subsystemColors = {}  # to store which subsystem gets which color
            for w in self.matroid.elems:
                r = cmod.getReaction(w)
                if (r != None) and (r.hasAnnotation("SUBSYSTEM")):
                    subsystem = r.getAnnotation("SUBSYSTEM")
                    if not subsystemColors.has_key(subsystem):
                        # create new color
                        c = generate_new_color(subsystemColors.values())
                        subsystemColors[subsystem] = c
                    c = subsystemColors[subsystem]
                    out += w.__str__() + " [color=\"%f %f %f\"]\n" % c
            # make legend
            out += "subgraph cluster_legend {\n"
            out += "label=\"Legend\";\n"
            colorcount = 0
            for (subsystem, color) in subsystemColors.items():
                out += "subsystemcolor%d" % colorcount
                out += " [label = \"%s\" " % subsystem
                out += "color = \"%f %f %f\"]\n" % color
                colorcount += 1
            out += "}\n"
        out += "}\n"
        return out

    def write(self, filename):
        """ writes this decomposition to file """
        fo = open(filename, mode='w')
        width = self.getWidth()
        variables = len(self.matroid.elems)
        rank = len(self.matroid.basis)
        fo.write("vars %d\n" % variables)
        fo.write("rank %d\n" % rank)
        fo.write("width %d\n" % width)
        found = set()
        for v in self.listNonLeaves():
            for (w, e) in v.edges.items():
                if w not in found:
                    fo.write("%s %s %d\n" % (v.label(), w.label(), len(e)))
            found.add(v)
        fo.close()


class NotFullyBranchedException(Exception):
    """ Thrown if the branch decomposition is not fully branched."""

    def __init__(self, vertex):
        self.vertex = vertex

    def __str__(self):
        return "%s is not fully branched." % self.vertex


class EdgeWidthException(Exception):

    def __init__(self, vertex1, vertex2, conn):
        self.vertex1 = vertex1
        self.vertex2 = vertex2
        self.conn = conn

    def __str__(self):
        out = "connectivity %d\n" % self.conn
        if isinstance(self.vertex1, Vertex):
            e1 = self.vertex1.edges[self.vertex2]
            out += "%s to %s has: %s\n" % (self.vertex1, self.vertex2, e1)
        if isinstance(self.vertex2, Vertex):
            e1 = self.vertex1.edges[self.vertex2]
            out += "%s to %s has: %s\n" % (self.vertex1, self.vertex2, e1)
        return out



"""
CBMPy: CBMultiEnv module
========================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBMultiEnv.py 710 2020-04-27 14:22:34Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

import os
import time
import numpy
from . import CBRead, CBWrite, CBTools, CBNetDB, CBSolver, CBMultiCore

HAVE_XLWT = False
try:
    import xlwt

    HAVE_XLWT = True
except:
    pass

from .CBConfig import __CBCONFIG__ as __CBCONFIG__

__DEBUG__ = __CBCONFIG__['DEBUG']
__version__ = __CBCONFIG__['VERSION']

# print('\n***\nMultiple Environment Module ({})\n***\n'.format(__version__))


def addProteinCostAnnotation(
    F, reaction_gene_map, gene_peptide_map, gene_annotation_key='GENE ASSOCIATION'
):
    reaction_gene_map = reaction_gene_map.copy()
    gene_peptide_map = gene_peptide_map.copy()
    for r in F.reactions:
        if reaction_gene_map[r.getId()] != None:
            r.annotation.update(
                {
                    'CBM_PEPTIDE_LENGTHS': [
                        (g, gene_peptide_map[g]) for g in reaction_gene_map[r.getId()]
                    ]
                }
            )
        else:
            r.annotation.update({'CBM_PEPTIDE_LENGTHS': []})
        GA = None
        if gene_annotation_key in r.annotation:
            GA = gene_annotation_key
        elif 'GENE ASSOCIATION' in r.annotation:
            GA = 'GENE ASSOCIATION'
        elif 'GENE_ASSOCIATION' in r.annotation:
            GA = 'GENE_ASSOCIATION'
        if GA != None:
            newAss = r.annotation[GA]
            for rl in r.annotation['CBM_PEPTIDE_LENGTHS']:
                # print rl
                # print newAss
                newAss = newAss.replace(rl[0], str(rl[1]))
            newAss = newAss.replace('and', '+').replace('(', '').replace(')', '')
            r.annotation.update({'CBM_PEPTIDE_SUBSTITUTED_LENGTHS': newAss})
            # print r.annotation
            if 'or' in r.annotation[GA]:
                # print ' '
                # print r.annotation[GA]
                # print r.annotation['CBM_PEPTIDE_SUBSTITUTED_LENGTHS']
                # print r.annotation['CBM_PEPTIDE_SUBSTITUTED_LENGTHS'].split('or')
                splitor = []
                for ass in r.annotation['CBM_PEPTIDE_SUBSTITUTED_LENGTHS'].split('or'):
                    exec('splitor.append(%s)' % ass)
                # print splitor
                r.annotation.update({'CBM_PEPTIDE_LENGTH_MIN': min(splitor)})
                r.annotation.update({'CBM_PEPTIDE_LENGTH_MAX': max(splitor)})
            else:
                if (
                    r.annotation['CBM_PEPTIDE_SUBSTITUTED_LENGTHS'] != None
                    and r.annotation['CBM_PEPTIDE_SUBSTITUTED_LENGTHS'] != ''
                ):
                    maxL = 0
                    # print r.annotation['CBM_PEPTIDE_SUBSTITUTED_LENGTHS']
                    print(r.annotation['CBM_PEPTIDE_SUBSTITUTED_LENGTHS'])
                    try:
                        exec(
                            'maxL = %s'
                            % r.annotation['CBM_PEPTIDE_SUBSTITUTED_LENGTHS']
                        )
                    except:
                        time.sleep(1)
                    r.annotation.update({'CBM_PEPTIDE_LENGTH_MIN': maxL})
                    r.annotation.update({'CBM_PEPTIDE_LENGTH_MAX': maxL})
                else:
                    r.annotation.update({'CBM_PEPTIDE_LENGTH_MIN': None})
                    r.annotation.update({'CBM_PEPTIDE_LENGTH_MAX': None})
            # print r.getId(), r.annotation['CBM_PEPTIDE_LENGTH_MIN'], r.annotation['CBM_PEPTIDE_LENGTH_MAX']
        else:
            r.annotation.update({'CBM_PEPTIDE_SUBSTITUTED_LENGTHS': None})
            r.annotation.update({'CBM_PEPTIDE_LENGTH_MIN': None})
            r.annotation.update({'CBM_PEPTIDE_LENGTH_MAX': None})


def scaleProteinCostAnnotation(fba):
    LEN_MAX = max(
        [
            r.annotation['CBM_PEPTIDE_LENGTH_MAX']
            for r in fba.reactions
            if r.annotation['CBM_PEPTIDE_LENGTH_MAX'] != None
        ]
    )
    LEN_AVG = numpy.average(
        [
            r.annotation['CBM_PEPTIDE_LENGTH_MAX']
            for r in fba.reactions
            if r.annotation['CBM_PEPTIDE_LENGTH_MAX'] != None
        ]
    )

    # cost function (AAlen/avg(AAlen) or 1.0
    for R in fba.reactions:
        if R.annotation['CBM_PEPTIDE_LENGTH_MAX'] == None:
            R.annotation['CBM_PEPTIDE_COST'] = 1.0
        else:
            R.annotation['CBM_PEPTIDE_COST'] = (
                R.annotation['CBM_PEPTIDE_LENGTH_MAX'] / LEN_AVG
            )
        R.annotation['CBM_AVG_PEPTIDE_LENGTH'] = LEN_AVG
    print('\nAverage protein length:', LEN_AVG)
    # print R.getId(), R.annotation['CBM_PEPTIDE_COST'], R.annotation['CBM_PEPTIDE_LENGTH_MAX']


# Standard amino acid abbreviations with ATP formation/production cost: Stouthamer (1973) Table 4 page 555
amino_acid_table = [
    ['A', 'Ala', 'Alanine', '+1'],
    ['R', 'Arg', 'Arginine', '-3'],
    ['N', 'Asn', 'Asparagine', '-2'],
    ['D', 'Asp', 'Aspartic acid', '+0'],
    ['C', 'Cys', 'Cysteine', '-3'],
    ['E', 'Glu', 'Glutamic acid', '+1'],
    ['Q', 'Gln', 'Glutamine', '+0'],
    ['G', 'Gly', 'Glycine', '+0'],
    ['H', 'His', 'Histidine', '-7'],
    ['I', 'Ile', 'Isoleucine', '-1'],
    ['L', 'Leu', 'Leucine', '+3'],
    ['K', 'Lys', 'Lysine', '+0'],
    ['M', 'Met', 'Methionine', '-4'],
    ['F', 'Phe', 'Phenylalanine', '-2'],
    ['P', 'Pro', 'Proline', '+0'],
    ['S', 'Ser', 'Serine', '+0'],
    ['T', 'Thr', 'Threonine', '-2'],
    ['W', 'Trp', 'Tryptophan', '-5'],
    ['Y', 'Tyr', 'Tyrosine', '-2'],
    ['V', 'Val', 'Valine', '+2'],
]


def runMultiStateFBA(
    input_list, dev_factor, options, dofva=False, fva_selected_reactions=None
):
    if options['algorithm'] == "ABS_L1_METHOD":
        if options['with_cost']:
            work_dir = os.path.join(options['base_work_dir'], 'AbsL1_cost')
        else:
            work_dir = os.path.join(options['base_work_dir'], 'AbsL1_nocost')
    elif options['algorithm'] == "L1_METHOD":
        if options['with_cost']:
            work_dir = os.path.join(options['base_work_dir'], 'L1_cost')
        else:
            work_dir = os.path.join(options['base_work_dir'], 'L1_nocost')

    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    IO_list = []
    ModelPrefixStrIO = ''
    for I in input_list:
        inputB = I[0]
        prefix = I[1]
        environ = None
        try:
            environ = I[2]
            print('INFO: custom environment defined:', environ)
        except:
            print('INFO: no custom environment defined')
        ModelPrefixStrIO += prefix
        ModelName = '%s(%s)' % (options['model_name_base'], inputB)

        fmg1 = CBRead.readSBML2FBA(options['input_model_file'], options['model_dir'])
        # get active (1st) objfunc
        biomass_name = fmg1.objectives[0].fluxObjectives[0].reaction
        # set name and inputs
        fmg1.id = prefix + options['model_name_base']
        # set environment
        if environ != None:
            environ = tuple(environ)
            for j_ in environ:
                print('Reaction', j_[0])
                print('LB', j_[1])
                print('UB', j_[2])
                assert fmg1.getReaction(j_[0]) != None, (
                    '\n%s is not a valid reaction id!' % j_[0]
                )
                fmg1.setReactionLowerBound(j_[0], j_[1])
                fmg1.setReactionUpperBound(j_[0], j_[2])
            # os.sys.exit()
        # prefix names
        fmg1.setPrefix(prefix, 'all')
        inputB = prefix + inputB
        # write LP
        CBTools.addStoichToFBAModel(fmg1)

        # cpx1 = CBSolver.cplx_constructLPfromFBA(fmg1, fname=None)
        LPFile1 = CBWrite.writeModelLP(fmg1, work_dir)
        cpx1 = CBSolver.cplx_getModelFromLP(LPFile1)
        CBSolver.cplx_setObjective(
            cpx1, prefix + 'obj1', [(1.0, inputB)], 'max', reset=True
        )
        CBSolver.cplx_setBounds(cpx1, inputB, min=-options['myInf'], max=0.0)
        CBSolver.cplx_setSingleConstraint(
            cpx1,
            'C_' + prefix + biomass_name,
            expr=[(1, prefix + biomass_name)],
            sense='E',
            rhs=options['target_biomass'],
        )
        CBSolver.cplx_writeLPtoLPTfile(
            cpx1,
            os.path.join(
                work_dir, prefix + '%s(%s)_Imin' % (options['model_name_base'], I[0])
            ),
            title='%s input minimization' % I[0],
            Dir=None,
        )
        CBSolver.cplx_Solve(cpx1)
        fba_sol1, objf_name1, objf_val1 = CBSolver.cplx_getOptimalSolution(cpx1)
        # for j in fba_sol:
        # print '%s: %s' % (j, fba_sol[j])
        print('\nCPLEX returns objective {} = {}'.format(objf_name1, objf_val1))
        slv_stat = CBSolver.cplx_getSolutionStatus(cpx1)
        print('\nStatus: {}'.format(slv_stat))
        IO_list.append((I[0], objf_val1, I[1], slv_stat, environ))

    print(
        '\nInput values for target: {} = {}'.format(
            biomass_name, options['target_biomass']
        )
    )
    F = file(
        os.path.join(
            work_dir,
            'find_input(%s)_(%s).csv' % (ModelPrefixStrIO, options['target_biomass']),
        ),
        'w',
    )
    F.write('input,value,prefix,status,environ\n')
    for IO in IO_list:
        CBTools.processBiGGchemFormula(fmg1)
        CBTools.processSBMLAnnotationNotes(fmg1, annotation_key='note')

        # print '%s, %.20f, %s, %s\n' % IO
        F.write('%s,%3.8f,%s,%s,\"%s\"\n' % IO)

    F.flush()
    F.close()

    DB = None
    if options['with_cost']:
        ##  db_cursor, db_conn = CBTools.ConnectDB('ecoli.sqlite', work_dir=options['DB_dir'])
        DB = CBNetDB.DBTools()
        DB.connectSQLiteDB('ecoli.sqlite', options['DB_dir'])

    multisolve = []
    for IO in IO_list:
        print(IO[3])
        assert IO[3] == 'LPS_OPT', '\nBAD LP ...'
        inputB = IO[0]
        inputValue = IO[1]
        environ = None
        try:
            print('Environ', IO[4])
            environ = IO[4]
            print('INFO: custom environment defined:', environ)
        except:
            print('INFO: no custom environment defined')
        # inputValue = round(IO[1], 8) # note this is a thumbsuck
        prefix = IO[2]
        fmg2 = CBRead.readSBML2FBA(
            options['input_model_file'], work_dir=options['model_dir']
        )
        # fix annotations
        CBTools.processBiGGchemFormula(fmg2)
        CBTools.processSBMLAnnotationNotes(fmg2, annotation_key='note')

        if options['with_cost']:
            # process GENE ANNOTATIONS USE initial version of model
            react_gene = CBTools.getModelGenesPerReaction(
                fmg2, gene_pattern=options['gene_pattern']
            )
            gene_react, ecoli_genes, no_gene = CBTools.getReactionsPerGene(react_gene)
            # recreate KeGG identifiers
            ecoli_genes = [options['model_KeGG_prefix'] + g for g in ecoli_genes]
            # retrieve and process amino acid sequences

            # extract info from database
            gene_peplen = {}
            for G in ecoli_genes:
                Glen = DB.fetchAll('SELECT aa_len FROM ecoli_genes WHERE gene="%s"' % G)
                if len(Glen) > 0 and len(Glen[0]) > 0:
                    Glen = Glen[0][0]
                    gene_peplen.update(
                        {G.replace(options['model_KeGG_prefix'], ''): Glen}
                    )
                else:
                    print('WARNING: no DB data for gene: {}'.format(G))
            ##  gene_peplen = CBTools.getPeptideLengthsFromDB(db_cursor, ecoli_genes, options['model_KeGG_prefix'])
            addProteinCostAnnotation(fmg2, react_gene, gene_peplen)
            scaleProteinCostAnnotation(fmg2)

        fmg2.setReactionBound(inputB, inputValue, 'lower')
        if options['relax_input_bound']:
            fmg2.setReactionBound(inputB, 0.0, 'upper')
        else:
            fmg2.setReactionBound(inputB, inputValue, 'upper')

        # set environment
        if environ != None:
            for j_ in environ:
                print('Reaction', j_[0])
                print('LB', j_[1])
                print('UB', j_[2])
                assert fmg2.getReaction(j_[0]) != None, (
                    '\n%s is not a valid reaction id!' % j_[0]
                )
                fmg2.setReactionLowerBound(j_[0], j_[1])
                fmg2.setReactionUpperBound(j_[0], j_[2])
            # raw_input('Im bored. What play thing can you offer me today?')

        # prefix names
        fmg2.id = prefix + options['model_name_base']
        fmg2.setPrefix(prefix, 'all')

        CBTools.addStoichToFBAModel(fmg2)

        fmg2lp = CBSolver.cplx_constructLPfromFBA(fmg2, fname=None)
        # LPFile2 = CBWrite.writeModelLP(fmg2, work_dir=work_dir)
        # fmg2lp = CBSolver.cplx_getModelFromLP(LPFile2)
        CBSolver.cplx_Solve(fmg2lp)
        (
            fmg2.objectives[0].solution,
            fmg2id,
            fmg2.objectives[0].value,
        ) = CBSolver.cplx_getOptimalSolution(fmg2lp)
        slv_stat = CBSolver.cplx_getSolutionStatus(fmg2lp)
        print('\nStatus: {}'.format(slv_stat))
        print("{} = {}".format(fmg2id, fmg2.objectives[0].value))
        multisolve.append(fmg2)
        time.sleep(3)

    # quick and dirty check to see that the reactions are in the same order
    if options['run_order_test']:
        for M in range(1, len(multisolve)):
            for j in range(len(multisolve[0].N.col)):
                # print multisolve[0].N.col[j][2:], multisolve[M].N.col[j][2:]
                assert multisolve[0].N.col[j][2:] == multisolve[M].N.col[j][2:], (
                    '\nN col mismatch: %s not e %s'
                    % (multisolve[0].N.col[j][2:], multisolve[M].N.col[j][2:])
                )
            for j in range(len(multisolve[0].N.row)):
                # print multisolve[0].N.row[j][2:], multisolve[M].N.row[j][2:]
                assert multisolve[0].N.row[j][2:] == multisolve[M].N.row[j][2:], (
                    '\nN row mismatch: %s != %s'
                    % (multisolve[0].N.row[j][2:], multisolve[M].N.row[j][2:])
                )
            idx = multisolve[0].getReactionIds()
            test_idx = multisolve[M].getReactionIds()
            for j in range(len(idx)):
                print(idx[j][2:], test_idx[j][2:])
                assert (
                    idx[j][2:] == test_idx[j][2:]
                ), '\nReaction ID mismatch: %s != %s' % (idx[j][2:], test_idx[j][2:])
        print('\nOrder test completed succesfully')
        os.sys.exit()

    # get model prefixes in various forms for output happiness
    ModelPrefixes = [m.prefix for m in multisolve]
    ModelPrefixStr = ''
    for p in ModelPrefixes:
        ModelPrefixStr += p
    print(ModelPrefixStr)

    objective_base = tuple([f.objectives[0].value for f in multisolve])

    RESULT_DATA = []
    RESULT_LPS = []

    ##  RESULT = numpy.zeros((len(GENERIC_REACTION_OBJECTS), len(multisolve)*len(dev_factor)+1), numpy.double)
    for devF in range(len(dev_factor)):
        for f in range(len(multisolve)):
            multisolve[f].objectives[0].value = objective_base[f] * dev_factor[devF]
        if options['with_cost']:
            if options['algorithm'] == "ABS_L1_METHOD":
                ##  raise NotImplementedError
                MultiFNbase = '%s_P(%s)_T(%s)_ABSL1_C' % (
                    options['input_model_file'],
                    ModelPrefixStr,
                    dev_factor[devF],
                )
                LPFileG = CBWrite.writeMinDistanceLP_absL1(
                    MultiFNbase,
                    multisolve,
                    work_dir=work_dir,
                    with_protein_cost=True,
                    ignoreDistance=[],
                    bigM=options['L1_BIGM_VALUE'],
                )
            elif options['algorithm'] == "L1_METHOD":
                MultiFNbase = '%s_P(%s)_T(%s)_L1_C' % (
                    options['input_model_file'],
                    ModelPrefixStr,
                    dev_factor[devF],
                )
                LPFileG = CBWrite.writeMinDistanceLP(
                    MultiFNbase, multisolve, work_dir=work_dir, with_protein_cost=True
                )
        else:
            if options['algorithm'] == "ABS_L1_METHOD":
                MultiFNbase = '%s_P(%s)_T(%s)_ABSL1' % (
                    options['input_model_file'],
                    ModelPrefixStr,
                    dev_factor[devF],
                )
                LPFileG = CBWrite.writeMinDistanceLP_absL1(
                    MultiFNbase,
                    multisolve,
                    work_dir=work_dir,
                    with_protein_cost=False,
                    ignoreDistance=[],
                    bigM=options['L1_BIGM_VALUE'],
                )
            elif options['algorithm'] == "L1_METHOD":
                MultiFNbase = '%s_P(%s)_T(%s)_L1' % (
                    options['input_model_file'],
                    ModelPrefixStr,
                    dev_factor[devF],
                )
                LPFileG = CBWrite.writeMinDistanceLP(
                    MultiFNbase, multisolve, work_dir=work_dir, with_protein_cost=False
                )

        mGlp = CBSolver.cplx_getModelFromLP(LPFileG)
        TIME_START = time.time()
        if options['algorithm'] == "ABS_L1_METHOD":
            # mGlp.parameters.threads.set(4)
            CBSolver.cplx_SolveMILP(mGlp, auto_mipgap=True)
        elif options['algorithm'] == "L1_METHOD":
            CBSolver.cplx_Solve(mGlp)
        TIME_END = time.time()

        print('\nSanity check')
        for f in multisolve:
            print('Model: {}'.format(f.id))
            for inp in input_list:
                print(
                    ' {} lower: {}'.format(
                        f.prefix + inp[0],
                        f.getFluxBoundByReactionID(f.prefix + inp[0], 'lower').value,
                    )
                )
                print(
                    ' {} upper: {}'.format(
                        f.prefix + inp[0],
                        f.getFluxBoundByReactionID(f.prefix + inp[0], 'upper').value,
                    )
                )

        print('\nFBA optima')
        for i in multisolve:
            print(' {} has optimum: {}'.format(i.id, i.objectives[0].value))

        mGsol, mGobid, mGobjval = CBSolver.cplx_getOptimalSolution(mGlp)

        print('\n*****\nMultipleInput(%s states) Result'.format(len(multisolve)))
        print(" optimum: %s = %s".format(mGobid, mGobjval))
        print(' status: %s'.format(CBSolver.cplx_getSolutionStatus(mGlp)))
        print('*****')

        print('\nMultipleInput({} states) Solve Time'.format(len(multisolve)))
        print(
            'Time: {}\nTime(m): {}\nTime(h): {}\n '.format(
                TIME_END - TIME_START,
                (TIME_END - TIME_START) / 60.0,
                (TIME_END - TIME_START) / 60.0 ** 2,
            )
        )

        GENERIC_REACTION_OBJECTS = multisolve[0].reactions
        GENERIC_PREFIX = multisolve[0].prefix
        try:
            FlxAO_J = numpy.zeros(
                (len(GENERIC_REACTION_OBJECTS), len(multisolve)), numpy.double
            )
            FlxAO_scaled = numpy.zeros(
                (len(GENERIC_REACTION_OBJECTS), len(multisolve) + 1), numpy.double
            )
        except AttributeError:
            FlxAO_J = numpy.zeros((len(GENERIC_REACTION_OBJECTS), len(multisolve)))
            FlxAO_scaled = numpy.zeros(
                (len(GENERIC_REACTION_OBJECTS), len(multisolve) + 1)
            )
        FluxNamesOut = []
        FluxScaling = []

        for flx in range(len(GENERIC_REACTION_OBJECTS)):
            Rid = GENERIC_REACTION_OBJECTS[flx].id[len(GENERIC_PREFIX) :]
            FluxNamesOut.append(Rid)
            if (
                options['with_cost']
                and 'CBM_PEPTIDE_COST' in GENERIC_REACTION_OBJECTS[flx].annotation
            ):
                cost = GENERIC_REACTION_OBJECTS[flx].annotation['CBM_PEPTIDE_COST']
            else:
                cost = 1.0
            FluxScaling.append(cost)
            for p in range(len(ModelPrefixes)):
                FlxAO_J[flx, p] = mGsol[ModelPrefixes[p] + Rid]
                FlxAO_scaled[flx, p] = mGsol[ModelPrefixes[p] + Rid] * cost
            # print Rid, FlxAO_scaled[flx]
        # print FlxAO_scaled.shape, ModelPrefixes
        # print FlxAO_scaled[:3]

        # TODO: AUTOMATE !!!!!
        # FLUX DIFFERENCE SUM
        all_flux_diff_xxx = numpy.zeros(3)
        if len(multisolve) == 2:
            all_flux_diff = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 1])
            all_flux_diff_xxx = numpy.abs(
                numpy.abs(FlxAO_scaled[:, 0]) - numpy.abs(FlxAO_scaled[:, 1])
            )
        elif len(multisolve) == 3:
            all_flux_diff1 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 1])
            all_flux_diff2 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 2])
            all_flux_diff3 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 2])
            all_flux_diff = all_flux_diff1 + all_flux_diff2 + all_flux_diff3
            all_flux_diff_xxx1 = numpy.abs(
                numpy.abs(FlxAO_scaled[:, 0]) - numpy.abs(FlxAO_scaled[:, 1])
            )
            all_flux_diff_xxx2 = numpy.abs(
                numpy.abs(FlxAO_scaled[:, 0]) - numpy.abs(FlxAO_scaled[:, 2])
            )
            all_flux_diff_xxx3 = numpy.abs(
                numpy.abs(FlxAO_scaled[:, 1]) - numpy.abs(FlxAO_scaled[:, 2])
            )
            all_flux_diff_xxx = (
                all_flux_diff_xxx1 + all_flux_diff_xxx2 + all_flux_diff_xxx3
            )
        elif len(multisolve) == 4:
            all_flux_diff1 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 1])
            all_flux_diff2 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 2])
            all_flux_diff3 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 3])
            all_flux_diff4 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 2])
            all_flux_diff5 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 3])
            all_flux_diff6 = numpy.abs(FlxAO_scaled[:, 2] - FlxAO_scaled[:, 3])
            all_flux_diff = (
                all_flux_diff1
                + all_flux_diff2
                + all_flux_diff3
                + all_flux_diff4
                + all_flux_diff5
                + all_flux_diff6
            )
        elif len(multisolve) == 5:
            all_flux_diff1 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 1])
            all_flux_diff2 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 2])
            all_flux_diff3 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 3])
            all_flux_diff4 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 4])
            all_flux_diff5 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 2])
            all_flux_diff6 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 3])
            all_flux_diff7 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 4])
            all_flux_diff8 = numpy.abs(FlxAO_scaled[:, 2] - FlxAO_scaled[:, 3])
            all_flux_diff9 = numpy.abs(FlxAO_scaled[:, 2] - FlxAO_scaled[:, 4])
            all_flux_diff10 = numpy.abs(FlxAO_scaled[:, 3] - FlxAO_scaled[:, 4])
            all_flux_diff = (
                all_flux_diff1
                + all_flux_diff2
                + all_flux_diff3
                + all_flux_diff4
                + all_flux_diff5
                + all_flux_diff6
                + all_flux_diff7
                + all_flux_diff8
                + all_flux_diff9
                + all_flux_diff10
            )
        elif len(multisolve) == 6:
            all_flux_diff1 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 1])
            all_flux_diff2 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 2])
            all_flux_diff3 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 3])
            all_flux_diff4 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 4])
            all_flux_diff5 = numpy.abs(FlxAO_scaled[:, 0] - FlxAO_scaled[:, 5])
            all_flux_diff6 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 2])
            all_flux_diff7 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 3])
            all_flux_diff8 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 4])
            all_flux_diff9 = numpy.abs(FlxAO_scaled[:, 1] - FlxAO_scaled[:, 5])
            all_flux_diff10 = numpy.abs(FlxAO_scaled[:, 2] - FlxAO_scaled[:, 3])
            all_flux_diff11 = numpy.abs(FlxAO_scaled[:, 2] - FlxAO_scaled[:, 4])
            all_flux_diff12 = numpy.abs(FlxAO_scaled[:, 2] - FlxAO_scaled[:, 5])
            all_flux_diff13 = numpy.abs(FlxAO_scaled[:, 3] - FlxAO_scaled[:, 4])
            all_flux_diff14 = numpy.abs(FlxAO_scaled[:, 3] - FlxAO_scaled[:, 5])
            all_flux_diff15 = numpy.abs(FlxAO_scaled[:, 4] - FlxAO_scaled[:, 5])
            all_flux_diff = (
                all_flux_diff1
                + all_flux_diff2
                + all_flux_diff3
                + all_flux_diff4
                + all_flux_diff5
                + all_flux_diff6
                + all_flux_diff7
                + all_flux_diff8
                + all_flux_diff9
                + all_flux_diff10
                + all_flux_diff11
                + all_flux_diff12
                + all_flux_diff13
                + all_flux_diff14
                + all_flux_diff15
            )
        else:
            all_flux_diff = FlxAO_scaled[:, -1]

        FlxAO_scaled[:, -1] = all_flux_diff

        print('Flux difference', all_flux_diff.sum(), all_flux_diff_xxx.sum())

        if options['algorithm'] == "ABS_L1_METHOD":
            SSAall = ({}, {}, {})
        elif options['algorithm'] == "L1_METHOD":
            SSAall = CBSolver.cplx_getSensitivities(mGlp)
            CBWrite.writeSensitivitiesToCSV(
                SSAall, os.path.join(work_dir, '%s.ssa.csv' % MultiFNbase)
            )

        DD = {
            'dev_factor': dev_factor[devF],
            'model_file': MultiFNbase,
            'solution_status': CBSolver.cplx_getSolutionStatus(mGlp),
            'obj_val': mGobjval,
            'obj_id': mGobid,
            'solution_cplex': mGsol.copy(),
            'solution': FlxAO_J,
            'solution_scaled': FlxAO_scaled,
            'flux_names': tuple(FluxNamesOut),
            'flux_diff': all_flux_diff.sum(),
            'flux_diff_new': all_flux_diff_xxx.sum(),
            'flux_scaling': tuple(FluxScaling),
            'sensitivity': SSAall,
            'prefixes': tuple(ModelPrefixes),
            'input_list': tuple(input_list),
            'multisolve': multisolve,
        }
        del SSAall
        RESULT_DATA.append(DD)
        RESULT_LPS.append(mGlp)

        FR = file(os.path.join(work_dir, '%s_rpt.txt' % MultiFNbase), 'w')
        FR.write('MultipleInput(%s states) Result:\n' % len(multisolve))
        # FR.write(' Flux diff: %s\n' % all_flux_diff.sum())
        FR.write(' optimum: %s = %s\n' % (mGobid, mGobjval))
        FR.write(' status: %s\n\n' % CBSolver.cplx_getSolutionStatus(mGlp))
        if options['algorithm'] == "ABS_L1_METHOD":
            FR.write('Distance: %s\n' % 'Sum(||J_a_x| - |J_b_x|| ...)')
        elif options['algorithm'] == "L1_METHOD":
            FR.write('Distance: %s\n' % 'sum(|J_a_x - J_b_x| ...)')
        FR.write('Big M: %s\n' % options['L1_BIGM_VALUE'])
        FR.write('With cost scaling: %s\n\n' % options['with_cost'])
        FR.write('MultipleInput(%s states) Solve Time\n' % len(multisolve))
        FR.write(
            'Time: %s\nTime(m): %s\nTime(h): %s\n\n'
            % (
                TIME_END - TIME_START,
                (TIME_END - TIME_START) / 60.0,
                (TIME_END - TIME_START) / 60.0 ** 2,
            )
        )

        FR.close()

        FD = file(os.path.join(work_dir, '%s_dat.test.csv' % MultiFNbase), 'w')
        head = 'Flux, '
        for p in ModelPrefixes:
            head += '%s,' % p
        FD.write('%sDiff,ProtCost,Diff(scaled)\n' % head)
        for F in range(FlxAO_scaled.shape[0]):
            FD.write('%s,' % FluxNamesOut[F])
            # for C in range(FlxAO_scaled.shape[1]):
            # if C != FlxAO_scaled.shape[1]-1:
            ##  FD.write('%s,' % FlxAO_scaled[F, C])
            # else:
            ##  FD.write('%s\n' % FlxAO_scaled[F, C])
            for C in range(FlxAO_scaled.shape[1]):
                FD.write('%s,' % FlxAO_scaled[F, C])
            ##  FD.write('%s,%s,%s\n' % (FluxScaling[F], FluxScaling[F]*FlxAO_scaled[F,-1]))
            FD.write('%s,%s\n' % (FluxScaling[F], FlxAO_scaled[F, -1]))
        FD.flush()
        FD.close()
        FD = file(os.path.join(work_dir, '%s_dat.csv' % MultiFNbase), 'w')
        head = 'Flux, '
        for p in ModelPrefixes:
            head += '%s,' % p
        FD.write('%sProtCost\n' % head)
        for F in range(FlxAO_J.shape[0]):
            FD.write('%s,' % FluxNamesOut[F])
            # for C in range(FlxAO_J.shape[1]):
            # if C != FlxAO_J.shape[1]-1:
            ##  FD.write('%s,' % FlxAO_J[F, C])
            # else:
            ##  FD.write('%s\n' % FlxAO_J[F, C])
            for C in range(FlxAO_J.shape[1]):
                FD.write('%s,' % FlxAO_J[F, C])
            FD.write('%s\n' % (FluxScaling[F]))
        FD.flush()
        FD.close()

    if options['with_cost']:
        OFname = '%s_P(%s)_C' % (options['input_model_file'], ModelPrefixStr)
    else:
        OFname = '%s_P(%s)' % (options['input_model_file'], ModelPrefixStr)
    if options['algorithm'] == "ABS_L1_METHOD":
        CBTools.storeObj(
            RESULT_DATA, os.path.join(work_dir, '%s_ABSL1_multi.dat' % OFname)
        )
    elif options['algorithm'] == "L1_METHOD":
        CBTools.storeObj(
            RESULT_DATA, os.path.join(work_dir, '%s_L1_multi.dat' % OFname)
        )

    # multi file
    # if options['algorithm'] == "ABS_L1_METHOD":
    # FD = file(os.path.join(work_dir, '%s_ABSL1_multi.csv' % OFname), 'w')
    # elif options['algorithm'] == "L1_METHOD":
    # FD = file(os.path.join(work_dir, '%s_L1_multi.csv' % OFname), 'w')

    # dat_arr = [a['solution'] for a in RESULT_DATA]
    # dat_arr = numpy.hstack(dat_arr)

    # hcntr = 0
    # h2cntr = 0
    # head = 'Flux, '
    # for p in (ModelPrefixes)*len(dev_factor):
    # if hcntr <= len(ModelPrefixes)-1:
    # head += '%s(%s%s),' % (p, '', dev_factor[h2cntr])
    # hcntr += 1
    # else:
    # h2cntr += 1
    # head += '%s(%s%s),' % (p, '', dev_factor[h2cntr])
    # hcntr = 0

    # print '\n'
    # FD.write('%s\n' % head[:-1])
    # for F in range(dat_arr.shape[0]):
    # FD.write('%s,' % FluxNamesOut[F])
    # for C in range(dat_arr.shape[1]):
    # if C != dat_arr.shape[1]-1:
    # FD.write('%s,' % dat_arr[F, C])
    # else:
    # FD.write('%s\n' % dat_arr[F, C])
    # del dat_arr

    # start multi output
    dat_arr = numpy.hstack([a['solution'] for a in RESULT_DATA])
    if options['algorithm'] == "ABS_L1_METHOD":
        FDNAME = os.path.join(work_dir, '%s_ABSL1_multi.csv' % OFname)

    elif options['algorithm'] == "L1_METHOD":
        FDNAME = os.path.join(work_dir, '%s_L1_multi.csv' % OFname)

    H = []
    for R in range(len(dev_factor)):
        for S in range(len(ModelPrefixes)):
            H.append('%s(%s%s)' % (ModelPrefixes[S], '', dev_factor[R]))
    print(H)
    print(dat_arr.shape)
    CBTools.exportLabelledArrayWithHeader2CSV(
        dat_arr, fname=FDNAME, names=FluxNamesOut, header=H
    )
    del dat_arr
    # end multi output

    opt_out = [(a['dev_factor'], a['obj_val'], a['flux_diff']) for a in RESULT_DATA]
    if options['algorithm'] == "ABS_L1_METHOD":
        FR2 = file(os.path.join(work_dir, '%s_ABSL1_opt.csv' % OFname), 'w')
    elif options['algorithm'] == "L1_METHOD":
        FR2 = file(os.path.join(work_dir, '%s_L1_opt.csv' % OFname), 'w')
    head = 'ObjDist(%s),optimum (%s),raw diff (%s)' % (
        ModelPrefixStr,
        ModelPrefixStr,
        ModelPrefixStr,
    )
    FR2.write('%s\n' % head)
    for d in opt_out:
        FR2.write('%s,%s,%s\n' % (d[0], d[1], d[2]))
    FR2.close()

    if dofva:
        MemDumpFiles = []
        for L in range(len(RESULT_LPS)):
            mdtempFile = os.path.join(work_dir, '_memDump_%s.lp' % L)
            RESULT_LPS[L].write(mdtempFile, filetype='lp')
            MemDumpFiles.append(mdtempFile)

        for L in range(len(RESULT_LPS) - 1, -1, -1):
            tmp = RESULT_LPS.pop(L)
            del tmp
        del RESULT_LPS

        print(MemDumpFiles)

        fva_time_start = time.time()
        fva_meta_res = []

        if HAVE_XLWT:
            wbk = xlwt.Workbook()
            wbk2 = xlwt.Workbook()
            wbk3 = xlwt.Workbook()

        GO = True
        pCntr = [1 for a in range(len(RESULT_DATA[L]['prefixes']))]
        for L in range(len(MemDumpFiles)):
            if HAVE_XLWT:
                dfStr = str(RESULT_DATA[L]['dev_factor'])
                sheet = wbk.add_sheet(dfStr)
                sheet.write(0, 0, 'id(%s)' % dfStr)
                sheet.write(0, 1, 'J(%s)' % dfStr)
                sheet.write(0, 2, 'min(%s)' % dfStr)
                sheet.write(0, 3, 'max(%s)' % dfStr)

            dLP = CBSolver.cplex.Cplex(MemDumpFiles[L])
            # remove artificial variables
            lp_var = dLP.variables.get_names()
            if fva_selected_reactions == None:
                lp_var = [j for j in lp_var if 'xvar_' not in j]
                lp_var = [j for j in lp_var if 'absL_' not in j]
                lp_var = [j for j in lp_var if 'zvar1' not in j]
            else:
                for r in fva_selected_reactions:
                    assert r in lp_var, 'invalid input to FVA reaction selection'
                lp_var = fva_selected_reactions
            pFx = RESULT_DATA[L]['prefixes']
            print(pFx)
            if HAVE_XLWT:
                if GO:
                    sheetL3 = []
                    for p_ in pFx:
                        s0 = wbk3.add_sheet(p_)
                        s0.write(0, 0, p_)
                        s0.write(0, 1, 'MaxBio pc(%s)' % p_.replace('_', ''))
                        s0.write(0, 2, 'opt(%s)' % p_.replace('_', ''))
                        s0.write(0, 3, 'min(%s)' % p_.replace('_', ''))
                        s0.write(0, 4, 'max(%s)' % p_.replace('_', ''))
                    for v_ in lp_var:
                        s1 = wbk2.add_sheet(v_)
                        s1.write(0, 0, v_)
                        s1.write(0, 1, 'opt')
                        s1.write(0, 2, 'min')
                        s1.write(0, 3, 'max')
                    GO = False

                    # g_R_EX_etoh_e	Opt(g)	Min(g)	Max(g)

            FVAtol = None
            FVAoptper = 100.0
            if 'FVAtol' in options:
                FVAtol = options['FVAtol']
            if 'FVAoptper' in options:
                FVAoptper = options['FVAoptper']

            if options['procs'] == None:
                print('\nOld single threaded algorithm\n')
                fva_res, fva_names = CBSolver.cplx_MultiFluxVariabilityAnalysis(
                    dLP,
                    selected_reactions=lp_var,
                    tol=FVAtol,
                    rhs_sense='upper',
                    optPercentage=FVAoptper,
                    work_dir=work_dir,
                    debug=False,
                )
            else:
                print(
                    '\nNew multi-threaded algorithm using {} threads.\n'.format(
                        options['procs']
                    )
                )
                fva_res, fva_names = CBMultiCore.runMultiCoreMultiEnvFVA(
                    dLP,
                    selected_reactions=lp_var,
                    tol=FVAtol,
                    rhs_sense='upper',
                    optPercentage=FVAoptper,
                    work_dir=work_dir,
                    debug=False,
                    procs=options['procs'],
                )
            fva_meta_res.append((fva_res, fva_names))
            FVAOname = '%s(%s)_fva.csv' % (OFname, RESULT_DATA[L]['dev_factor'])
            FVAOname = os.path.join(work_dir, FVAOname)
            CBTools.exportLabelledArrayWithHeader(
                fva_res,
                fname=FVAOname,
                fva_names=fva_names,
                header=['Opt', 'Min', 'Max', 'MinS', 'MaxS'],
                sep=',',
                format='%f',
            )
            if HAVE_XLWT:
                rfact = 8
                for j_ in range(1, len(fva_names) + 1):
                    # ex 1
                    J = fva_names[j_ - 1]
                    sheet.write(j_, 0, J)
                    # ex 2
                    sheet2 = wbk2.get_sheet(j_ - 1)
                    sheet2.write(L + 1, 0, RESULT_DATA[L]['dev_factor'])
                    sheet2.write(L + 1, 1, fva_res[j_ - 1, 0])
                    res = fva_res[j_ - 1, 1]
                    if numpy.isnan(res):
                        # res = fva_res[j_-1, 0]
                        res = 'NaN'
                    else:
                        res = round(res, rfact)
                    sheet2.write(L + 1, 2, res)
                    res = fva_res[j_ - 1, 2]
                    if numpy.isnan(res):
                        # res = fva_res[j_-1, 0]
                        res = 'NaN'
                    else:
                        res = round(res, rfact)
                    sheet2.write(L + 1, 3, res)
                    for d_ in range(1, 4):
                        res = fva_res[j_ - 1, d_ - 1]
                        if numpy.isnan(res):
                            # res = fva_res[j_-1, 0]
                            res = 'NaN'
                        else:
                            res = round(res, rfact)
                        sheet.write(j_, d_, res)
                    # ex 3

                    for p_ in range(len(pFx)):
                        if J.startswith(pFx[p_]):
                            sheet3 = wbk3.get_sheet(p_)
                            # wbk3.save('fuckme.xls')
                            sheet3.write(pCntr[p_], 0, J)
                            sheet3.write(pCntr[p_], 1, RESULT_DATA[L]['dev_factor'])
                            sheet3.write(pCntr[p_], 2, fva_res[j_ - 1, 0])
                            res = fva_res[j_ - 1, 1]
                            if numpy.isnan(res):
                                # res = fva_res[j_-1, 0]
                                res = 'NaN'
                            else:
                                res = round(res, rfact)
                            sheet3.write(pCntr[p_], 3, res)
                            res = fva_res[j_ - 1, 2]
                            if numpy.isnan(res):
                                # res = fva_res[j_-1, 0]
                                res = 'NaN'
                            else:
                                res = round(res, rfact)
                            sheet3.write(pCntr[p_], 4, res)
                            pCntr[p_] = pCntr[p_] + 1
            del dLP
        if HAVE_XLWT:
            t = str(time.time()).split('.')[0][-5:]
            xlname = 'data_%s_(%s)' % (
                OFname.replace('.xml', '').replace('base', ''),
                t,
            )
            wbk.save(os.path.join(work_dir, xlname + '.fva_per_objdist.xls'))
            wbk2.save(os.path.join(work_dir, xlname + '.fva_per_flux.xls'))
            wbk3.save(os.path.join(work_dir, xlname + '.fva_per_input.xls'))
        fva_time_end = time.time()
        print(
            '\n*****\nTotal ({} LP\'s) time for FVA ({} threads) min: {}\n*****\n'.format(
                len(MemDumpFiles) * len(lp_var),
                options['procs'],
                (fva_time_end - fva_time_start) / 60.0,
            )
        )
        F = file(os.path.join(work_dir, '%s_fva_stats.txt' % OFname), 'a')
        F.write(
            '*****\nTotal (%s LP\'s) time for FVA (%s threads) min: %s\n*****\n'
            % (
                len(MemDumpFiles) * len(lp_var),
                options['procs'],
                (fva_time_end - fva_time_start) / 60.0,
            )
        )
        F.close()
        CBTools.storeObj(
            fva_meta_res, os.path.join(work_dir, '%s_fva_meta.dat' % OFname)
        )
    return True
